#!/usr/bin/perl -w

# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl Fb.t'

#########################

# change 'tests => 1' to 'tests => last_test_to_print';

use strict;
use warnings;

use Data::Dumper;
use IO::String;

use Test::More tests => 165;

BEGIN {
	use_ok('Fb::Out');
	use_ok('Fb::Network');
	use_ok('Fb::Dialog');
	use_ok('Fb::IO');
	use_ok('Fb::Data');
	use_ok('Fb::Cache');
};

use Fb::Data qw(listEquality hashEquality setEquality leq in min max hashify limitWords diffStrings wrapLine splitObjects bash_quote calcResult timeStamp $MAXINT64 align compactArray retireBackups getXmlXc);
use Fb::IO qw(readFile iaexec $iaeexitcode readDir readXmlFile);
use Fb::Dialog qw(input menu);
use Fb::Out qw(renderHelp getCommandBlock);
use Fb::Cache;

sub TestData {
	my $prefix="t";
	# Test set-handling
	ok( setEquality({a => 1,b =>0},{a => 0,b => 1}) == 1,"SetEq with differing values");
	ok( setEquality({a => 1},{b => 1}) == 0,"SetEq with differing keys");

	ok( hashEquality({a => 1,b => 2},{a => 1,b => 2}) == 1,"HashEq");
	ok( hashEquality({a => 1,b => 2},{a => 0,b => 1}) == 0,"HashEq with differing values");
	ok( hashEquality({a => 1,b => 2},{a => 1,c => 2}) == 0,"HashEq with differing keys");
	ok( hashEquality({a => 1,b => 2},{a => 1}) == 0,"HashEq with differing # of keys");

	ok( listEquality([1,2,3],[1,2,3]) == 1,"ListEq");
	ok( listEquality([1,2,3],[1,2]) == 0,"ListEq with length()-diff");
	ok( listEquality([1,2,3],[3,2,1]) == 0,"ListEq with differing content");
	ok( in([1,2,3],1) == 1,"'in'-match");
	ok( in([1,2,3],4) == 0,"'in', no match");
	ok( max(1,2,3,4,5) == 5,"max");
	ok( max(1,5,3,4,2) == 5,"max unsorted");
	ok( ! defined(max()),"max with empty list");
	ok( min(5,4,3,2,1) == 1,"min");
	ok( min(5,1,3,2,4) == 1,"min unsorted");
	ok( ! defined(min()),"min with empty list");

	ok( bash_quote("a\\\"!|\$` ;&<>()[]*#a'\n") eq ('a\\\\\"\!\|\$\`\ \;\&\<\>\(\)\[\]\*\#a' . "\\'\$'\\n'"), "bash_quote");
	ok( bash_quote('') eq '""', "bash_quote empty");
	ok( bash_quote(undef) eq '', "bash_quote undef");
	
	ok( limitWords([1,2,3],3) eq "1 2 3","limitWords, words shown");
	ok( limitWords([1,2,3,4],3) eq "[4]","limitWords, count shown");
	
	ok( hashEquality({hashify('a')},{a => 1}) == 1,"hashify - scalar");
	ok( hashEquality({hashify(undef)},{}) == 1,"hashify - undef");
	ok( hashEquality({hashify(['a','b'])},{a => 1,b => 1}) == 1,"hashify - array");
	ok( hashEquality({hashify({'a' => 1,'b' => 1})},{a => 1,b => 1}) == 1,"hashify - hash");

	my $diffstring;
	ok( diffStrings("a\nb\n","a\nb\n",getdiff => \$diffstring) == 1,"diffStrings - eq");
	ok( $diffstring eq '',"diffStrings - eq text");
	ok( diffStrings("a\n","a\nb\n",getdiff => \$diffstring) == 0,"diffStrings - add");
	ok( $diffstring eq "> b\n","diffStrings - add text");
	ok( diffStrings("a\nb\n","a\nc\n",getdiff => \$diffstring) == 0,"diffStrings - mod");
	ok( $diffstring eq "< b\n> c\n","diffStrings - mod text");
	ok( diffStrings(['a','b'],['a','b'],getdiff => \$diffstring) == 1,"diffStrings ARRAY - eq ");
	ok( $diffstring eq '',"diffStrings ARRAY - eq text");
	ok( diffStrings(['a','b'],['a','c'],getdiff => \$diffstring) == 0,"diffStrings ARRAY - mod");
	ok( $diffstring eq "< b\n> c\n","diffStrings ARRAY - mod text");

	my @short=qw(a b c);
	ok( wrapLine(\@short,5) eq "a b c\n","wrapLine one line");
	ok( wrapLine(\@short,1) eq "a\nb\nc\n","wrapLine three lines");
	ok( wrapLine(\@short,5,prefix => 'x') eq "xa b\nxc\n","wrapLine prefixed");
	ok( wrapLine(\@short,5,abfprefix => 'x') eq "a b\nxc\n","wrapLine abfprefixed");
	
	ok( listEquality([splitObjects(['a','b c','d,e'])],['a','b','c','d','e']),"splitObjects 1");
	ok( hashEquality({splitObjects(['a','b'],format => 'hash')},{a => 1,b => 1}),"splitObjects 2");
	ok( listEquality(splitObjects(['a','b c','d,e'],format => 'arrayref'),['a','b','c','d','e']),"splitObjects 3");
	ok( hashEquality(splitObjects(['a','b'],format => 'hashref'),{a => 1,b => 1}),"splitObjects 4");

	my @mr_map=( # $mainresult,$newresult,$newmainresult,$return
		[0,0,0,0],[0,1,0,0],[0,2,2,0],[0,3,2,0],
		[1,0,0,0],[1,1,1,1],[1,2,2,0],[1,3,3,1],
		[2,0,2,0],[2,1,2,0],[2,2,2,0],[2,3,2,0],
		[3,0,2,0],[3,1,3,1],[3,2,2,0],[3,3,3,1],
	);
	for my $tuple (@mr_map) {
		my $mr=$tuple->[0];
		my $return=calcResult(\$mr,$tuple->[1]);
		ok ( $mr == $tuple->[2],"calcResult [$tuple->[0],$tuple->[1],mainresult:$mr!=$tuple->[2]!]");
		ok ( $return == $tuple->[3],"calcResult [$tuple->[0],$tuple->[1],return:$return!=$tuple->[3]]");
	}
	my $time=1408775389;
	ok( timeStamp('YYYYMMDD-hhmmss',$time) eq '20140823-082949',"timeStamp YYYYMMDD-hhmmss");
	ok( timeStamp('---',$MAXINT64) eq '-',"timeStamp infinite");
	ok( timeStamp('---',$MAXINT64/2) eq '?',"timeStamp invalid");
	ok( timeStamp('YYYY-MM-DD hh:mm:ss',$time) eq '2014-08-23 08:29:49',"timeStamp YYYY-MM-DD hh:mm:ss");
	ok( timeStamp('YYYYMMDD-hhmm',$time) eq '20140823-0829',"timeStamp YYYYMMDD-hhmm");
	ok( timeStamp('DD.MM.YYYY',$time) eq '23.08.2014',"timeStamp DD.MM.YYYY");
	ok( timeStamp('YYYYMMDD',$time) eq '20140823',"timeStamp YYYYMMDD");

	ok( align('test',6) eq 'test  ',"Align left +2");
	ok( align('test',4) eq 'test',"Align left fit");
	ok( align('test',2) eq 'te',"Align left -2");
	ok( align('test',-6) eq '  test',"Align right +2");
	ok( align('test',4) eq 'test',"Align right fit");
	ok( align('test',2) eq 'te',"Align right -2");
	ok( align('test',0) eq 'test',"Align nop");

	ok( listEquality(['1','2','3'],compactArray(['1','2','3'])), "compactArray 1");
	ok( listEquality(['2x1','2','2x3'],compactArray(['1','1','2','3','3'])), "compactArray 2");

	# Test backup retiring
	my $now=1560251101; # 2019-06-11 13:05:01
	my %existing_backups=(
		1560251101 => 'now, 2019-06-11 13:05:01',
		1560251100 => 'just before now. 2019-06-11 13:05:00',
	
		1560247502 => "< one hour ago, 2019-06-11 12:05:02",
		1560247501 => "exactly one hour ago, 2019-06-11 12:05:01",
		1560247200 => "~ one hour ago, 2019-06-11 12:00:00",
		1560243600 => "2 hours ago, 2019-06-11 11:00:00",
		1560240000 => "3 hours ago, 2019-06-11 10:00:00",
		1560236400 => "4 hours ago, 2019-06-11 09:00:00",
		1560232800 => "5 hours ago, 2019-06-11 08:00:00",
		
		1560164702 => "< one day ago, 2019-06-10 13:05:02",
		1560164701 => "exactly one day ago, 2019-06-10 13:05:01",
		1560164400 => "~ one day ago, 2019-06-10 13:00:00",
		1560078000 => "2 days ago, 2019-06-09 13:00:00",
		1559991600 => "3 days ago, 2019-06-08 13:00:00",
		1559905200 => "4 days ago, 2019-06-07 13:00:00",
		1559818800 => "5 days ago, 2019-06-06 13:00:00",

		1559646000 => "~ one week ago, 2019-06-04 13:00:00",
		1559041200 => "2 weeks ago, 2019-05-28 13:00:00",
		1559041200 => "3 weeks ago, 2019-05-28 13:00:00",
		1557226800 => "5 weeks ago, '2019-05-07 13:00:00",
		
		1557572400 => "~ one month ago, 2019-05-11 13:00:00",
		1554980400 => "2 months ago, 2019-04-11 13:00:00",
		1552305600 => "3 months ago, 2019-03-11 13:00:00",
		1549886400 => "4 months ago, 2019-02-11 13:00:00",
		1547208000 => "5 months ago, 2019-01-11 13:00:00",
	);
	my %backup_test=(
		'0,0,0,0' => [1560251101],
		'1,0,0,0' => [1560251101,1560247200],
		'0,1,0,0' => [1560251101,1560164400],
		'0,0,1,0' => [1560251101,1559646000],
		'0,0,0,1' => [1560251101,1557226800],
		'1,1,1,1' => [1557226800,1559646000,1560164400,1560247200,1560251101],
		'2,2,2,2' => [1554980400,1559041200,1560078000,1560164400,1560243600,1560247200,1560251101],
		'3,3,3,3' => [1560251101,1560247200,1560243600,1560240000,1560164400,1560078000,1559991600,1559041200,1554980400,1552305600],
	);
	my $tbd; # The hashref that will contain backups "to be deleted"
	for my $test (sort keys %backup_test) {
		my @retention_array=split /,/,$test;
		$tbd=retireBackups(\%existing_backups,\@retention_array,now => $now);
		my @reference_keep=sort @{$backup_test{$test}};
		my @calculated_keep=grep { ! $tbd->{$_} } sort keys %existing_backups;
		ok( listEquality(\@reference_keep,\@calculated_keep), "retireBackups($test)");
	}
	my $xml=readFile("$prefix/test.xml",format => 'string');
	my $xc=getXmlXc($xml);
	my $node=$xc->find('/test/toast[@id="test"]');
	ok( int($node->[0]->textContent() eq 'yes'),"getXmlXc");

	ok( leq(undef,undef) ,"leq(undef,undef)");
	ok( leq(1,"1") ,"leq(1,'1')");
	ok( ! leq(undef,1) ,"!leq(undef,x)");
	ok( leq(1,1,1) ,"leq(multi)");
	ok( ! leq(1,1,2) ,"!leq(multi)");
}

sub TestIO {
	my $result;
	my $prefix="t";
	$result=iaexec(["$prefix/iaexec-success"],'exec');
	ok( $result == 1,"iaexec exec success");
	$result=iaexec(["$prefix/iaexec-fail"],'exec');
	ok( $result == 0,"iaexec exec fail");
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => 'string');
	ok( $result eq "a\nb\nc\n","iaexec readpipe 1");
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => 'arrayref');
	ok( listEquality($result,['a','b','c']),"iaexec readpipe 2");
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => 'hashref');
	ok( hashEquality($result,{a => 1,b => 1,c => 1}),"iaexec readpipe 3");
	my $output;my @output;my %output;
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => \$output);
	ok( $result == 1,"iaexec readpipe 4 result");
	ok( $output eq "a\nb\nc\n","iaexec readpipe 4 output");
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => \@output);
	ok( $result == 1,"iaexec readpipe 5 result");
	ok( listEquality(\@output,['a','b','c']),"iaexec readpipe 5 output");
	$result=iaexec(["$prefix/iaexec-readpipe"],'readpipe',format => \%output);
	ok( $result == 1,"iaexec readpipe 6 result");
	ok( hashEquality(\%output,{a => 1,b => 1,c => 1}),"iaexec readpipe 7 output");

	# Test submitting quoted parameters to commands
	$result=iaexec(["$prefix/iaexec-params",'$ (){}\"*`?|&<>!;'],'exec');
	ok( $result == 1,'iaexec quoted params');
	
	$result=iaexec("$prefix/iaexec-params " . bash_quote('$ (){}\"*`?|&<>!;'),'exec');
	ok( $result == 1,'iaexec quoted params 2');
	
	# Test bash expansions
	$ENV{'TEST'}='$ (){}\"*`?|&<>!;';
	$result=iaexec("$prefix/iaexec-params \"\$TEST\"",'exec');
	ok( $result == 1,'iaexec unquoted params');

	# exec with vector parameters should not by mistake call the binary without space
	$result=iaexec(["$prefix/iaexec-with space"],'test',failure => [1,3]);
	ok( ($iaeexitcode >> 8) == 2,'iaexec with spaced binary');
	
	$result=iaexec("$prefix/iaexec-with space",'test',failure => [1,2]);
	ok( ($iaeexitcode >> 8) == 3,"iaexec without spaced binary $iaeexitcode");

	$result=iaexec(["$prefix/iaexec-withstderr",0,'test'],'readpipe',stderr => 'include',format => \@output);
	ok( listEquality(\@output,['stdout','stderr']),"iaexec with included stderr (execv)");
	$result=iaexec("$prefix/iaexec-withstderr",'readpipe',stderr => 'include',format => \@output);
	ok( listEquality(\@output,['stdout','stderr']),"iaexec with included stderr (single)");

	$result=iaexec("$prefix/iaexec-readpipe",'exec',stdout => 'redirect:_tempfile');
	my $tempcontent=readFile('_tempfile',format => 'string');
	ok( $tempcontent eq "a\nb\nc\n","iaexec with redirect");
	unlink('_tempfile');

	$result=iaexec(["$prefix/iaexec-withstderr",5],'readpipe',stderr => 'redirect:_tempfile',format => \$output,failure => [6]);
	$tempcontent=readFile('_tempfile',format => 'string');
	ok( $output eq "stdout\n","iaexec with redirect split 1/3");
	ok( $tempcontent eq "stderr\n","iaexec with redirect split 2/3");
	ok( $result == 1,"iaexec with redirect split 3/3");
	unlink('_tempfile');

	$result=iaexec(["$prefix/iaexec-writepipe"],'writepipe',input => "a\nb\n");
	ok( $result == 1,"iaexec writepipe with string input");
	$result=iaexec(["$prefix/iaexec-writepipe"],'writepipe',input => "b\na\n");
	ok( $result == 0,"iaexec writepipe fail");
	$result=iaexec(["$prefix/iaexec-writepipe"],'writepipe',input => ["a","b"]);
	ok( $result == 1,"iaexec writepipe with arrayref input");

	$result=readFile("$prefix/readfile",format => 'string');
	ok( $result eq "a\nb\nc\n","readFile string");
	$result=readFile("$prefix/readfile-nofnl",format => 'string');
	ok( $result eq "a\nb\nc","readFile string nofnl");
	$result=readFile("$prefix/readfile",format => 'arrayref');
	ok( listEquality($result,['a','b','c']),"readFile arrayref");
	$result=readFile("$prefix/readfile",format => 'hashref');
	ok( hashEquality($result,{a => 1,b => 1,c => 1}),"readFile hashref");

	# Unicode/binary
	$result=readFile("$prefix/readfile-utf8",format => 'string');
	ok( $result eq "a\xc3\xa4o\xc3\xb6u\xc3\xbc\n","readFile binary");
	$result=readFile("$prefix/readfile-utf8",format => 'string',encoding => 'utf8');
	ok( $result eq "a\xe4o\xf6u\xfc\n","readFile UTF-8");
	writeFile("$prefix/writefile-utf8","a\xe4o\xf6u\xfc\n",encoding => 'utf8');
	$result= ! system("/usr/bin/cmp","$prefix/readfile-utf8","$prefix/writefile-utf8");
	ok( $result,"writeFile UTF-8");
	unlink("$prefix/writefile-utf8");

	# Readdir
	$result=readDir("$prefix/readdir",format => 'arrayref');
	$result=[grep { $_ ne '.svn' } @{$result}];
	ok( listEquality($result,['1','1a','2']),"readDir array unfiltered");
	$result=readDir("$prefix/readdir",filter => '1.*',format => 'arrayref');
	ok( listEquality($result,['1','1a']),"readDir array filtered");
	$result=readDir("$prefix/readdir",format => 'hashref');
	delete $result->{'.svn'};
	ok( hashEquality($result,{'1' => 1,'1a' => 1,'2' => 1}),"readDir hashref unfiltered");

	my %cache=();
	my $xc=readXmlFile("$prefix/test.xml",cache => \%cache);
	my $node=$xc->find('/test/toast[@id="test"]');
	ok( int($node->[0]->textContent() eq 'yes'),"readXmlFile");
	
}

sub iaexec_fake_dialog {
	my ($command,$mode,$param)=@_;
	my $fh;
	if ( $command eq "dialog --output-fd 3 3>&1 >&2 --title testtitle --inputbox testtext\\ \$'\\n'second\\ line 24 80 defaulttext" ) {
		$$param=IO::String->new("nondefaulttext\n");
	} elsif ( $command eq 'dialog --output-fd 3 3>&1 >&2 --title cancel --inputbox testtext 24 80 ""' ) {
		$$param=IO::String->new("");
		return 1 << 8;
	} elsif ( $command eq 'dialog --output-fd 3 3>&1 >&2 --title title --no-cancel --ok-label select --menu text 10 80 3 1 entry\ 1 2 2 ' ) {
		$$param=IO::String->new("1\n");
	} elsif ( $command eq 'dialog --output-fd 3 3>&1 >&2 --title cancel --no-cancel --ok-label select --menu text 10 80 3 1 entry\ 1 2 2 ' ) {
		$$param=IO::String->new("");
		return 1 << 8;
	} elsif ( $command eq 'dialog --output-fd 3 3>&1 >&2 --title title --cancel-label cancel --help-button --help-label select --ok-label ignore --menu text 10 80 3 1 entry\ 1 2 2 ' ) {
		$$param=IO::String->new("HELP 2\n");
		return 2 << 8;
	} elsif ( $command eq 'dialog --output-fd 3 3>&1 >&2 --title title --cancel-label cancel --help-button --help-label ignore2 --extra-button --extra-label select --ok-label ignore --menu text 10 80 3 1 entry\ 1 2 2 ') {
		$$param=IO::String->new("2\n");
		return 3 << 8;
	} else {
		out('e',"Unexpected call while faking iaexec()");
		out('i',"Command: " . $command);
	}
	$?=0;
	return 0;
}

sub TestOut {
	my $help=0;my $dummy=0;
	my %commands=(
		commandline => "libfb-perl-test {[options]} [parameters] [...]",
		author => "Frank Burkhardt <burk\@cbs.mpg.de>",
		digest => "Just a functional test",
		blocks => [
			{
				title => "Commands",
				parameters => [
					{name => 'help|h',var => \$help,text => ["Help page"]},
					{name => 'version|V',var => \$dummy,text => ["Show program version"]},
				],
			},
			{
				title => "Options",
				parameters => [
					{name => 'someoption|O',var => \$dummy,text => ["Some option"]},
					{name => 'someoption2|o=s',var => \$dummy,arg => 'PAR',text => [
						"Some option with a parameter. Possible parameters PAR are:",
						['a - some parameter','b - some other parameter','c - another one'],
					]},
					{name => 'someoption3|x=s',var => \$dummy,arg => 'PAR',text => [
						"Some option with a parameter. Possible parameters PAR:",
						{
							'a' => 'one of the best parameter ever',
							'b' => 'the second best parameter',
							'c' => 'completely useless parameter',
						}
					]},
				],
			},
			getCommandBlock(ftest => 1),
		],
		examples => [
			{command => "libfb-perl-test -O -o a",
				text => "Do something useful"},
			{command => "libfb-perl-test --someoption3 c",
				text => "So something else"},
		],
	);
	my $text1=renderHelp(%commands,test => 1);
	my $text2=readFile("t/help-test.txt",format => 'string');
	ok( $text1 eq $text2,"renderHelp()");
}

sub TestDialog {
	Fb::IO::prepareFtests(iaexec => \&iaexec_fake_dialog);
	$Fb::Dialog::DIALOG_MAXX=80;
	$Fb::Dialog::DIALOG_MAXY=24;

	# Test, if the dialog calls are translated into correct 'dialog' commandlines
	ok( input("testtitle","testtext \nsecond line","defaulttext") eq 'nondefaulttext',"Dialog::input()");
	ok( ! defined(input("cancel","testtext")),"Dialog::input() canceled");

	ok( listEquality([menu('title','text',[['1','entry 1'],['2','2']],['select'],size_x => 80,size_y => 10)],[1,'1']),"Dialog::menu() default" );
	ok( listEquality([menu('cancel','text',[['1','entry 1'],['2','2']],['select'],size_x => 80,size_y => 10)],[0,undef]),"Dialog::menu() cancel" );
	ok( listEquality([menu('title','text',[['1','entry 1'],['2','2']],['ignore','cancel','select'],size_x => 80,size_y => 10)],[2,'2']),"Dialog::menu() custom1" );
	ok( listEquality([menu('title','text',[['1','entry 1'],['2','2']],['ignore','cancel','ignore2','select'],size_x => 80,size_y => 10)],[3,'2']),"Dialog::menu() custom2" );
}

sub TestCache {
	my $cache=Fb::Cache->new();
	my $value1=1;
	my $value2=sub { 1+1 };
	ok( leq($cache->set('test',$value1),$value1),"Cache::set(simple) returns value");
	ok( leq($cache->get('test'),$value1),"Cache::get(simple)");
	
	$cache->set('test',$value2);
	ok( leq($cache->get('test'),2),"Cache::set(replacement)");
	
	ok( leq($cache->get('test2'),undef),"Cache::get(empty)");
	ok( leq($cache->get('test2',create => $value1),$value1),"Cache::get(with_default_value)");
	ok( leq($cache->get('test3',create => $value2),2),"Cache::get(with_default_sub)");

	ok( listEquality([sort @{$cache->keys()}],['test','test2','test3']),"Cache::keys1");
	ok( $cache->removePattern('test[23]') == 2,"Cache::removePattern");
	ok( listEquality([sort @{$cache->keys()}],['test']),"Cache::keys2/removePattern");
	$cache->remove('test');
	ok( listEquality([sort @{$cache->keys()}],[]),"Cache::remove");
}

TestData();
TestIO();
TestOut();
TestDialog();
TestCache();
